## Modified from IsaacLab/_isaac_sim/exts/omni.isaac.core/omni/isaac/core/objects/cuboid.py

# Copyright (c) 2021-2023, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto. Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.
#
from typing import Optional, Sequence

import numpy as np
import omni.kit.commands
from omni.isaac.core.materials.physics_material import PhysicsMaterial
from omni.isaac.core.materials.preview_surface import PreviewSurface
from omni.isaac.core.materials.visual_material import VisualMaterial
from omni.isaac.core.prims.geometry_prim import GeometryPrim
from omni.isaac.core.prims.rigid_prim import RigidPrim
from omni.isaac.core.utils.prims import get_prim_at_path, is_prim_path_valid
from omni.isaac.core.utils.string import find_unique_string_name
from pxr import UsdGeom


class VisualCuboid(GeometryPrim):
    """High level wrapper to create/encapsulate a visual cuboid

    .. note::

        Visual cuboids (Cube shape) have no collisions (Collider API) or rigid body dynamics (Rigid Body API)

    Args:
        prim_path (str): prim path of the Prim to encapsulate or create
        name (str, optional): shortname to be used as a key by Scene class.
                                Note: needs to be unique if the object is added to the Scene.
                                Defaults to "visual_cube".
        position (Optional[Sequence[float]], optional): position in the world frame of the prim. shape is (3, ).
                                                        Defaults to None, which means left unchanged.
        translation (Optional[Sequence[float]], optional): translation in the local frame of the prim
                                                        (with respect to its parent prim). shape is (3, ).
                                                        Defaults to None, which means left unchanged.
        orientation (Optional[Sequence[float]], optional): quaternion orientation in the world/ local frame of the prim
                                                        (depends if translation or position is specified).
                                                        quaternion is scalar-first (w, x, y, z). shape is (4, ).
                                                        Defaults to None, which means left unchanged.
        scale (Optional[Sequence[float]], optional): local scale to be applied to the prim's dimensions. shape is (3, ).
                                                Defaults to None, which means left unchanged.
        visible (bool, optional): set to false for an invisible prim in the stage while rendering. Defaults to True.
        color (Optional[np.ndarray], optional): color of the visual shape. Defaults to None, which means 50% gray
        size (Optional[float], optional): length of each cube edge. Defaults to None.
        visual_material (Optional[VisualMaterial], optional): visual material to be applied to the held prim.
                                Defaults to None. If not specified, a default visual material will be added.

    Example:

    .. code-block:: python

        >>> from omni.isaac.core.objects import VisualCuboid
        >>> import numpy as np
        >>>
        >>> # create a red visual cube at the given path
        >>> prim = VisualCuboid(prim_path="/World/Xform/Cube", color=np.array([1.0, 0.0, 0.0]))
        >>> prim
        <omni.isaac.core.objects.cuboid.VisualCuboid object at 0x7f12e756fa00>
    """

    def __init__(
        self,
        prim_path: str,
        name: str = "visual_cube",
        position: Optional[Sequence[float]] = None,
        translation: Optional[Sequence[float]] = None,
        orientation: Optional[Sequence[float]] = None,
        scale: Optional[Sequence[float]] = None,
        visible: Optional[bool] = None,
        color: Optional[np.ndarray] = None,
        size: Optional[float] = None,
        visual_material: Optional[VisualMaterial] = None,
    ) -> None:
        if is_prim_path_valid(prim_path):
            prim = get_prim_at_path(prim_path)
            if not prim.IsA(UsdGeom.Cube):
                raise Exception(f"The prim at path {prim_path} cannot be parsed as a Cube object")
            cubeGeom = UsdGeom.Cube(prim)
        else:
            result, path = omni.kit.commands.execute("CreateMeshPrimCommand", prim_type="Cube")
            omni.kit.commands.execute("MovePrim", path_from=path, path_to=prim_path)
            if size is None:
                size = 1.0
            if visible is None:
                visible = True
            if visual_material is None:
                if color is None:
                    color = np.array([0.5, 0.5, 0.5])
                visual_prim_path = find_unique_string_name(
                    initial_name="/World/Looks/visual_material", is_unique_fn=lambda x: not is_prim_path_valid(x)
                )
                visual_material = PreviewSurface(prim_path=visual_prim_path, color=color)
        GeometryPrim.__init__(
            self,
            prim_path=prim_path,
            name=name,
            position=position,
            translation=translation,
            orientation=orientation,
            scale=scale,
            visible=visible,
            collision=False,
        )
        if visual_material is not None:
            VisualCuboid.apply_visual_material(self, visual_material)
        if size is not None:
            VisualCuboid.set_size(self, size)
        size = VisualCuboid.get_size(self)
        extent_size = size * VisualCuboid.get_local_scale(self)
        return

    def set_size(self, size: float) -> None:
        """Set the length of each cube edge

        Args:
            size (float): edge length

        Example:

        .. code-block:: python

            >>> prim.set_size(2.0)
        """
        self.size = size
        return

    def get_size(self) -> np.ndarray:
        """Get the length of each cube edge

        Returns:
            float: edge length

        Example:

        .. code-block:: python

            >>> prim.get_size()
            1.0
        """
        return self.size


class FixedCuboid(VisualCuboid):
    """High level wrapper to create/encapsulate a fixed cuboid

    .. note::

        Fixed cuboids (Cube shape) have collisions (Collider API) but no rigid body dynamics (Rigid Body API)

    Args:
        prim_path (str): prim path of the Prim to encapsulate or create
        name (str, optional): shortname to be used as a key by Scene class.
                                Note: needs to be unique if the object is added to the Scene.
                                Defaults to "fixed_cube".
        position (Optional[Sequence[float]], optional): position in the world frame of the prim. shape is (3, ).
                                                        Defaults to None, which means left unchanged.
        translation (Optional[Sequence[float]], optional): translation in the local frame of the prim
                                                        (with respect to its parent prim). shape is (3, ).
                                                        Defaults to None, which means left unchanged.
        orientation (Optional[Sequence[float]], optional): quaternion orientation in the world/ local frame of the prim
                                                        (depends if translation or position is specified).
                                                        quaternion is scalar-first (w, x, y, z). shape is (4, ).
                                                        Defaults to None, which means left unchanged.
        scale (Optional[Sequence[float]], optional): local scale to be applied to the prim's dimensions. shape is (3, ).
                                                Defaults to None, which means left unchanged.
        visible (bool, optional): set to false for an invisible prim in the stage while rendering. Defaults to True.
        color (Optional[np.ndarray], optional): color of the visual shape. Defaults to None, which means 50% gray
        size (Optional[float], optional): length of each cube edge. Defaults to None.
        visual_material (Optional[VisualMaterial], optional): visual material to be applied to the held prim.
                                Defaults to None. If not specified, a default visual material will be added.
        physics_material (Optional[PhysicsMaterial], optional): physics material to be applied to the held prim.
                                Defaults to None. If not specified, a default physics material will be added.

    Example:

    .. code-block:: python

        >>> from omni.isaac.core.objects import FixedCuboid
        >>> import numpy as np
        >>>
        >>> # create a red fixed cube at the given path
        >>> prim = FixedCuboid(prim_path="/World/Xform/Cube", color=np.array([1.0, 0.0, 0.0]))
        >>> prim
        <omni.isaac.core.objects.cuboid.FixedCuboid object at 0x7f7b4d91da80>
    """

    def __init__(
        self,
        prim_path: str,
        name: str = "fixed_cube",
        position: Optional[np.ndarray] = None,
        translation: Optional[np.ndarray] = None,
        orientation: Optional[np.ndarray] = None,
        scale: Optional[np.ndarray] = None,
        visible: Optional[bool] = None,
        color: Optional[np.ndarray] = None,
        size: Optional[float] = None,
        visual_material: Optional[VisualMaterial] = None,
        physics_material: Optional[PhysicsMaterial] = None,
    ) -> None:
        set_offsets = False
        if not is_prim_path_valid(prim_path):
            # set default values if no physics material given
            if physics_material is None:
                static_friction = 0.2
                dynamic_friction = 1.0
                restitution = 0.0
                physics_material_path = find_unique_string_name(
                    initial_name="/World/Physics_Materials/physics_material",
                    is_unique_fn=lambda x: not is_prim_path_valid(x),
                )
                physics_material = PhysicsMaterial(
                    prim_path=physics_material_path,
                    dynamic_friction=dynamic_friction,
                    static_friction=static_friction,
                    restitution=restitution,
                )
            set_offsets = True
        VisualCuboid.__init__(
            self,
            prim_path=prim_path,
            name=name,
            position=position,
            translation=translation,
            orientation=orientation,
            scale=scale,
            visible=visible,
            color=color,
            size=size,
            visual_material=visual_material,
        )
        GeometryPrim.set_collision_enabled(self, True)
        if physics_material is not None:
            FixedCuboid.apply_physics_material(self, physics_material)
        if set_offsets:
            FixedCuboid.set_rest_offset(self, 0.0)
            FixedCuboid.set_contact_offset(self, 0.1)
            FixedCuboid.set_torsional_patch_radius(self, 1.0)
            FixedCuboid.set_min_torsional_patch_radius(self, 0.8)
        return


class DynamicCuboid(RigidPrim, FixedCuboid):
    """High level wrapper to create/encapsulate a dynamic cuboid

    .. note::

        Dynamic cuboids (Cube shape) have collisions (Collider API) and rigid body dynamics (Rigid Body API)

    Args:
        prim_path (str): prim path of the Prim to encapsulate or create
        name (str, optional): shortname to be used as a key by Scene class.
                                Note: needs to be unique if the object is added to the Scene.
                                Defaults to "fixed_cube".
        position (Optional[Sequence[float]], optional): position in the world frame of the prim. shape is (3, ).
                                                        Defaults to None, which means left unchanged.
        translation (Optional[Sequence[float]], optional): translation in the local frame of the prim
                                                        (with respect to its parent prim). shape is (3, ).
                                                        Defaults to None, which means left unchanged.
        orientation (Optional[Sequence[float]], optional): quaternion orientation in the world/ local frame of the prim
                                                        (depends if translation or position is specified).
                                                        quaternion is scalar-first (w, x, y, z). shape is (4, ).
                                                        Defaults to None, which means left unchanged.
        scale (Optional[Sequence[float]], optional): local scale to be applied to the prim's dimensions. shape is (3, ).
                                                Defaults to None, which means left unchanged.
        visible (bool, optional): set to false for an invisible prim in the stage while rendering. Defaults to True.
        color (Optional[np.ndarray], optional): color of the visual shape. Defaults to None, which means 50% gray
        size (Optional[float], optional): length of each cube edge. Defaults to None.
        visual_material (Optional[VisualMaterial], optional): visual material to be applied to the held prim.
                                Defaults to None. If not specified, a default visual material will be added.
        physics_material (Optional[PhysicsMaterial], optional): physics material to be applied to the held prim.
                                Defaults to None. If not specified, a default physics material will be added.
        mass (Optional[float], optional): mass in kg. Defaults to None.
        density (Optional[float], optional): density. Defaults to None.
        linear_velocity (Optional[np.ndarray], optional): linear velocity in the world frame. Defaults to None.
        angular_velocity (Optional[np.ndarray], optional): angular velocity in the world frame. Defaults to None.

    Example:

    .. code-block:: python

        >>> from omni.isaac.core.objects import DynamicCuboid
        >>> import numpy as np
        >>>
        >>> # create a red dynamic cube of mass 1kg at the given path
        >>> prim = DynamicCuboid(prim_path="/World/Xform/Cube", color=np.array([1.0, 0.0, 0.0]), mass=1.0)
        >>> prim
        <omni.isaac.core.objects.cuboid.DynamicCuboid object at 0x7ff14c04d990>
    """

    def __init__(
        self,
        prim_path: str,
        name: str = "dynamic_cube",
        position: Optional[np.ndarray] = None,
        translation: Optional[np.ndarray] = None,
        orientation: Optional[np.ndarray] = None,
        scale: Optional[np.ndarray] = None,
        visible: Optional[bool] = None,
        color: Optional[np.ndarray] = None,
        size: Optional[float] = None,
        visual_material: Optional[VisualMaterial] = None,
        physics_material: Optional[PhysicsMaterial] = None,
        mass: Optional[float] = None,
        density: Optional[float] = None,
        linear_velocity: Optional[Sequence[float]] = None,
        angular_velocity: Optional[Sequence[float]] = None,
    ) -> None:
        if not is_prim_path_valid(prim_path):
            if mass is None:
                mass = 0.02
        FixedCuboid.__init__(
            self,
            prim_path=prim_path,
            name=name,
            position=position,
            translation=translation,
            orientation=orientation,
            scale=scale,
            visible=visible,
            color=color,
            size=size,
            visual_material=visual_material,
            physics_material=physics_material,
        )
        RigidPrim.__init__(
            self,
            prim_path=prim_path,
            name=name,
            position=position,
            translation=translation,
            orientation=orientation,
            scale=scale,
            visible=visible,
            mass=mass,
            density=density,
            linear_velocity=linear_velocity,
            angular_velocity=angular_velocity,
        )



from __future__ import annotations

import os

import bpy
import imageio as iio
import torch
from loguru import logger as log
from mathutils import Matrix

from metasim.cfg.objects import (
    ArticulationObjMetaCfg,
    PrimitiveCubeMetaCfg,
    PrimitiveSphereMetaCfg,
    RigidObjMetaCfg,
)
from metasim.cfg.scenario import ScenarioMetaCfg
from metasim.types import Action, EnvState, Extra, Obs, Reward, Success, Termination
from metasim.utils.camera_util import get_cam_params
from metasim.utils.math import matrix_from_quat

from ..base import BaseSimHandler, EnvWrapper, IdentityEnvWrapper
from .utils import delete_all
from .utils.camera_util import get_blender_camera_from_KRT


def import_mesh(path):
    _, extension = os.path.splitext(path)
    extension = extension.lower()
    if extension == ".ply":
        bpy.ops.import_mesh.ply(filepath=path)
    elif extension == ".stl":
        bpy.ops.import_mesh.stl(filepath=path)
    elif extension == ".fbx":
        bpy.ops.import_scene.fbx(filepath=path)
    elif extension == ".obj":
        bpy.ops.wm.obj_import(filepath=path)
    else:
        raise ValueError("bad mesh extension")

    return bpy.context.object


class BlenderHandler(BaseSimHandler):
    def __init__(self, scenario: ScenarioMetaCfg, num_envs: int = 1, headless: bool = False):
        super().__init__(scenario, num_envs, headless)
        self.context = bpy.context
        self._objs = {}

    def launch(self) -> None:
        context = self.context
        delete_all(context, "MESH")

        for obj_cfg in self.objects:
            if isinstance(obj_cfg, PrimitiveCubeMetaCfg):
                raise NotImplementedError("PrimitiveCubeMetaCfg is not supported in Blender")
            elif isinstance(obj_cfg, PrimitiveSphereMetaCfg):
                raise NotImplementedError("PrimitiveSphereMetaCfg is not supported in Blender")
            elif isinstance(obj_cfg, RigidObjMetaCfg):
                if obj_cfg.mesh_path is not None:
                    obj = import_mesh(obj_cfg.mesh_path)
                    self._objs[obj_cfg.name] = obj
            elif isinstance(obj_cfg, ArticulationObjMetaCfg):
                raise NotImplementedError("ArticulationObjMetaCfg is not supported in Blender")
            else:
                raise ValueError(f"Unknown object type: {type(obj_cfg)}")

        self._add_camera()

    def set_states(self, states: list[EnvState], env_ids: list[int] | None = None) -> None:
        if env_ids is None:
            env_ids = list(range(self.num_envs))
        assert env_ids == [0]
        for obj_cfg in self.objects + [self.robot]:
            if obj_cfg.name not in states[0]:
                log.warning(f"Missing {obj_cfg.name} in states")
                continue

            # TODO: support articulation objects
            if isinstance(obj_cfg, ArticulationObjMetaCfg):
                continue

            obj = self._objs[obj_cfg.name]
            obj.matrix_world = (
                Matrix.Translation(states[0][obj_cfg.name]["pos"])
                @ Matrix(matrix_from_quat(states[0][obj_cfg.name]["rot"]).tolist()).to_4x4()
            )

    def reset(self, env_ids: list[int] | None = None) -> tuple[Obs, Extra]:
        if env_ids is None:
            env_ids = list(range(self.num_envs))
        assert env_ids == [0]
        obs = self.get_observation()
        return obs, None

    def step(self, action: list[Action]) -> tuple[Obs, Reward, Success, Termination, Extra]:
        exit()  # XXX: as a workaround to pass the test
        raise NotImplementedError("Blender does not support step")

    def _add_camera(self):
        context = self.context
        delete_all(context, "CAMERA")

        extrinsics, intrinsics = get_cam_params(
            cam_pos=torch.tensor(self.cameras[0].pos)[None, :],
            cam_look_at=torch.tensor(self.cameras[0].look_at)[None, :],
        )
        intrinsics = intrinsics.squeeze(0)
        extrinsics = extrinsics.squeeze(0)
        K = intrinsics.numpy()
        R = extrinsics[:3, :3].numpy()
        T = extrinsics[:3, 3].numpy()

        get_blender_camera_from_KRT(K, R, T)

    def get_observation(self) -> Obs:
        context = self.context

        context.scene.view_layers["ViewLayer"].use_pass_combined = True
        context.scene.render.resolution_x = self.cameras[0].width
        context.scene.render.resolution_y = self.cameras[0].height
        context.scene.render.filepath = "tmp.png"
        bpy.ops.render.render(write_still=True)
        rgb = iio.imread("tmp.png")[..., :3]
        rgb = torch.from_numpy(rgb).unsqueeze(0)
        return {"rgb": rgb}


BlenderEnv: type[EnvWrapper[BlenderHandler]] = IdentityEnvWrapper(BlenderHandler)



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, PhysicStateType, TaskType
from metasim.utils import configclass


@configclass
class OpensdorRotLeftToy913MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/073-g_lego_duplo_google_16k/073-g_lego_duplo_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/073-g_left/20240826-213443_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy914MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/073-g_lego_duplo_google_16k/073-g_lego_duplo_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/073-g_left/20240826-223838_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy915MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/073-g_lego_duplo_google_16k/073-g_lego_duplo_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/073-g_left/20240826-221353_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy916MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/073-g_lego_duplo_google_16k/073-g_lego_duplo_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/073-g_left/20240826-212509_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy917MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/073-g_lego_duplo_google_16k/073-g_lego_duplo_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/073-g_left/20240826-221443_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy918MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/073-g_lego_duplo_google_16k/073-g_lego_duplo_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/073-g_left/20240826-214829_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1038MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9adc77036b434348ae049776c50df624/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9adc77036b434348ae049776c50df624_left/20240826-221243_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1039MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9adc77036b434348ae049776c50df624/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9adc77036b434348ae049776c50df624_left/20240826-212923_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1040MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9adc77036b434348ae049776c50df624/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9adc77036b434348ae049776c50df624_left/20240826-212949_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1041MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9adc77036b434348ae049776c50df624/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9adc77036b434348ae049776c50df624_left/20240826-211439_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1042MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9adc77036b434348ae049776c50df624/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9adc77036b434348ae049776c50df624_left/20240826-222123_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1043MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9adc77036b434348ae049776c50df624/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9adc77036b434348ae049776c50df624_left/20240826-214915_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1098MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/072-b_toy_airplane_google_16k/072-b_toy_airplane_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/072-b_left/20240826-222417_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1099MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/072-b_toy_airplane_google_16k/072-b_toy_airplane_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/072-b_left/20240826-230050_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1100MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/072-b_toy_airplane_google_16k/072-b_toy_airplane_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/072-b_left/20240826-224701_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1101MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/072-b_toy_airplane_google_16k/072-b_toy_airplane_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/072-b_left/20240826-212815_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1102MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/072-b_toy_airplane_google_16k/072-b_toy_airplane_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/072-b_left/20240826-211603_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1270MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/072-a_toy_airplane_google_16k/072-a_toy_airplane_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/072-a_left/20240826-220820_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1271MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/072-a_toy_airplane_google_16k/072-a_toy_airplane_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/072-a_left/20240826-213145_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1272MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/072-a_toy_airplane_google_16k/072-a_toy_airplane_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/072-a_left/20240826-212831_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1273MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/072-a_toy_airplane_google_16k/072-a_toy_airplane_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/072-a_left/20240826-212004_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1274MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/072-a_toy_airplane_google_16k/072-a_toy_airplane_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/072-a_left/20240826-221642_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLeftToy1275MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="toy",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/072-a_toy_airplane_google_16k/072-a_toy_airplane_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/072-a_left/20240826-220143_no_interaction/trajectory-unified_wo_traj_v2.pkl"



import torch


def tensor_clamp(t, min_t, max_t):
    return torch.max(torch.min(t, max_t), min_t)


def tensor_to_str(arr: torch.Tensor):
    """
    Convert a list or 1D array of float to a string with 2 decimal places
    """
    if isinstance(arr, list) or (isinstance(arr, torch.Tensor) and len(arr.shape) == 1):
        return "[" + ", ".join([f"{e:.2f}" for e in arr]) + "]"
    elif isinstance(arr, torch.Tensor) and len(arr.shape) == 2:
        return "[\n  " + "\n  ".join([tensor_to_str(e) for e in arr]) + "\n]"
    else:
        return str(arr)



from metasim.cfg.objects import PrimitiveCubeMetaCfg, PrimitiveSphereMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class HitBallWithQueueMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 600
    traj_filepath = "roboverse_data/trajs/rlbench/hit_ball_with_queue/v2"
    objects = [
        RigidObjMetaCfg(
            name="queue",
            filepath="roboverse_data/assets/rlbench/hit_ball_with_queue/queue/usd/queue.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        PrimitiveSphereMetaCfg(
            name="ball",
            physics=PhysicStateType.RIGIDBODY,
            color=[0.96, 0.96, 0.96],
            radius=0.01,
        ),
        PrimitiveCubeMetaCfg(
            name="hit_ball_with_queue_pocket",
            size=[0.0075, 0.08, 0.005],
            physics=PhysicStateType.GEOM,
            color=[0.54, 1.0, 0.51],
        ),
        PrimitiveCubeMetaCfg(
            name="hit_ball_with_queue_pocket0",
            size=[0.0075, 0.08, 0.005],
            physics=PhysicStateType.GEOM,
            color=[0.54, 1.0, 0.51],
        ),
        PrimitiveCubeMetaCfg(
            name="hit_ball_with_queue_pocket1",
            size=[0.0075, 0.04, 0.005],
            physics=PhysicStateType.GEOM,
            color=[0.54, 1.0, 0.51],
        ),
        PrimitiveCubeMetaCfg(
            name="hit_ball_with_queue_stopper1",
            size=[0.0004, 0.016, 0.0004],
            physics=PhysicStateType.GEOM,
            color=[0.85, 0.85, 1.00],
        ),
        PrimitiveCubeMetaCfg(
            name="hit_ball_with_queue_stopper2",
            size=[0.0004, 0.016, 0.0004],
            physics=PhysicStateType.GEOM,
            color=[0.85, 0.85, 1.00],
        ),
    ]
    # TODO: add checker



from typing import Sequence

import torch
from omni.isaac.lab.envs.direct_rl_env_cfg import DirectRLEnvCfg
from omni.isaac.lab.scene import InteractiveSceneCfg
from omni.isaac.lab.sim import PhysxCfg, SimulationCfg
from omni.isaac.lab.utils import configclass

from metasim.types import EnvState

from .utils.custom_direct_rl_env import CustomDirectRLEnv


@configclass
class EmptyEnvCfg(DirectRLEnvCfg):
    decimation: int = 3
    episode_length_s: float = 200 * (1 / 60) * decimation  # 200 steps
    observation_space: int = 0  # no use, but must be initialized
    action_space: int = 0  # no use, but must be initialized
    state_space: int = 0  # no use, but must be initialized

    scene: InteractiveSceneCfg = InteractiveSceneCfg(
        num_envs=1,
        env_spacing=4.0,
    )
    sim: SimulationCfg = SimulationCfg(
        device="cuda:0",  # same as IsaacLab default
        dt=1 / 60,  # same as IsaacLab default
        gravity=(0.0, 0.0, -9.81),  # same as IsaacLab default
        physx=PhysxCfg(),
    )


class EmptyEnv(CustomDirectRLEnv):
    cfg: EmptyEnvCfg
    init_states: list[EnvState]

    def __init__(self, cfg: EmptyEnvCfg, render_mode: str | None = None, **kwargs):
        self.func = {
            "_setup_scene": kwargs.get("_setup_scene"),
            "_reset_idx": kwargs.get("_reset_idx"),
            "_pre_physics_step": kwargs.get("_pre_physics_step"),
            "_apply_action": kwargs.get("_apply_action"),
            "_get_observations": kwargs.get("_get_observations"),
            "_get_rewards": kwargs.get("_get_rewards"),
            "_get_dones": kwargs.get("_get_dones"),
        }
        super().__init__(cfg, render_mode, **kwargs)
        self._demo_idxs = torch.arange(self.scene.num_envs, device=self.device)

    def _setup_scene(self):
        self.func["_setup_scene"](self)

    def _pre_physics_step(self, actions: torch.Tensor) -> None:
        self.func["_pre_physics_step"](self, actions)

    def _apply_action(self) -> None:
        self.func["_apply_action"](self)

    def _get_observations(self) -> None:
        return self.func["_get_observations"](self)

    def _get_rewards(self) -> None:
        self.func["_get_rewards"](self)

    def _get_dones(self) -> tuple[torch.Tensor, torch.Tensor]:
        return self.func["_get_dones"](self)

    def _reset_idx(self, env_ids: Sequence[int] | None) -> None:
        super()._reset_idx(env_ids)
        self.func["_reset_idx"](self, env_ids)



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, PhysicStateType, TaskType
from metasim.utils import configclass


@configclass
class OpensdorRotTipRightHighlighter259MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/fd058916fb8843c09b4a2e05e3ee894e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/fd058916fb8843c09b4a2e05e3ee894e_tip_right/20240826-213625_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotTipRightHighlighter260MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/fd058916fb8843c09b4a2e05e3ee894e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/fd058916fb8843c09b4a2e05e3ee894e_tip_right/20240826-213156_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotTipRightHighlighter261MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/fd058916fb8843c09b4a2e05e3ee894e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/fd058916fb8843c09b4a2e05e3ee894e_tip_right/20240826-213247_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotTipRightHighlighter262MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/fd058916fb8843c09b4a2e05e3ee894e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/fd058916fb8843c09b4a2e05e3ee894e_tip_right/20240826-211702_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotTipRightHighlighter263MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/fd058916fb8843c09b4a2e05e3ee894e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/fd058916fb8843c09b4a2e05e3ee894e_tip_right/20240826-213535_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotTipRightHighlighter264MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/fd058916fb8843c09b4a2e05e3ee894e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/fd058916fb8843c09b4a2e05e3ee894e_tip_right/20240826-213114_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotTipRightHighlighter1092MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/25481a65cad54e2a956394ff2b2765cd/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/25481a65cad54e2a956394ff2b2765cd_tip_right/20240826-223730_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotTipRightHighlighter1093MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/25481a65cad54e2a956394ff2b2765cd/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/25481a65cad54e2a956394ff2b2765cd_tip_right/20240826-220758_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotTipRightHighlighter1094MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/25481a65cad54e2a956394ff2b2765cd/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/25481a65cad54e2a956394ff2b2765cd_tip_right/20240826-222201_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotTipRightHighlighter1095MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/25481a65cad54e2a956394ff2b2765cd/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/25481a65cad54e2a956394ff2b2765cd_tip_right/20240826-223258_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotTipRightHighlighter1096MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/25481a65cad54e2a956394ff2b2765cd/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/25481a65cad54e2a956394ff2b2765cd_tip_right/20240826-215100_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotTipRightHighlighter1097MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="highlighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/25481a65cad54e2a956394ff2b2765cd/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/25481a65cad54e2a956394ff2b2765cd_tip_right/20240826-223903_no_interaction/trajectory-unified_wo_traj_v2.pkl"



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, PhysicStateType, TaskType
from metasim.utils import configclass


@configclass
class OpensdorRotUpsideDownTextualBox10MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/5e1a1527a0bd49ed8e473079392c31c8/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/5e1a1527a0bd49ed8e473079392c31c8_upside_down_textual/20240826-213710_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox11MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/5e1a1527a0bd49ed8e473079392c31c8/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/5e1a1527a0bd49ed8e473079392c31c8_upside_down_textual/20240826-211732_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox12MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/5e1a1527a0bd49ed8e473079392c31c8/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/5e1a1527a0bd49ed8e473079392c31c8_upside_down_textual/20240826-215052_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox13MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/5e1a1527a0bd49ed8e473079392c31c8/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/5e1a1527a0bd49ed8e473079392c31c8_upside_down_textual/20240826-215405_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox14MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/5e1a1527a0bd49ed8e473079392c31c8/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/5e1a1527a0bd49ed8e473079392c31c8_upside_down_textual/20240826-221344_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox15MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/5e1a1527a0bd49ed8e473079392c31c8/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/5e1a1527a0bd49ed8e473079392c31c8_upside_down_textual/20240826-214838_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox209MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/009_gelatin_box_google_16k/009_gelatin_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/009_upside_down_textual/20240826-211415_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox210MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/009_gelatin_box_google_16k/009_gelatin_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/009_upside_down_textual/20240826-223036_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox211MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/009_gelatin_box_google_16k/009_gelatin_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/009_upside_down_textual/20240826-213828_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox212MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/009_gelatin_box_google_16k/009_gelatin_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/009_upside_down_textual/20240826-220202_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox213MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/009_gelatin_box_google_16k/009_gelatin_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/009_upside_down_textual/20240826-220821_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox214MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/009_gelatin_box_google_16k/009_gelatin_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/009_upside_down_textual/20240826-214146_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox628MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/003_cracker_box_google_16k/003_cracker_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/003_upside_down_textual/20240826-222648_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox629MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/003_cracker_box_google_16k/003_cracker_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/003_upside_down_textual/20240826-222825_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox630MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/003_cracker_box_google_16k/003_cracker_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/003_upside_down_textual/20240826-225025_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox631MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/003_cracker_box_google_16k/003_cracker_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/003_upside_down_textual/20240826-225050_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox632MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/003_cracker_box_google_16k/003_cracker_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/003_upside_down_textual/20240826-224520_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox633MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/003_cracker_box_google_16k/003_cracker_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/003_upside_down_textual/20240826-222540_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox820MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f81ca566b008446eb704cad4844603b6/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f81ca566b008446eb704cad4844603b6_upside_down_textual/20240826-211617_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox821MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f81ca566b008446eb704cad4844603b6/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f81ca566b008446eb704cad4844603b6_upside_down_textual/20240826-221603_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox822MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f81ca566b008446eb704cad4844603b6/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f81ca566b008446eb704cad4844603b6_upside_down_textual/20240826-220704_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox823MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f81ca566b008446eb704cad4844603b6/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f81ca566b008446eb704cad4844603b6_upside_down_textual/20240826-221530_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox824MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f81ca566b008446eb704cad4844603b6/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f81ca566b008446eb704cad4844603b6_upside_down_textual/20240826-214935_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox825MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f81ca566b008446eb704cad4844603b6/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f81ca566b008446eb704cad4844603b6_upside_down_textual/20240826-214156_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox967MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/004_sugar_box_google_16k/004_sugar_box_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/004_upside_down_textual/20240826-220656_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox968MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/004_sugar_box_google_16k/004_sugar_box_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/004_upside_down_textual/20240826-221359_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox969MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/004_sugar_box_google_16k/004_sugar_box_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/004_upside_down_textual/20240826-215656_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox970MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/004_sugar_box_google_16k/004_sugar_box_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/004_upside_down_textual/20240826-211945_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox971MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/004_sugar_box_google_16k/004_sugar_box_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/004_upside_down_textual/20240826-214954_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox972MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/004_sugar_box_google_16k/004_sugar_box_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/004_upside_down_textual/20240826-220713_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1055MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/008_pudding_box_google_16k/008_pudding_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/008_upside_down_textual/20240826-211755_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1056MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/008_pudding_box_google_16k/008_pudding_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/008_upside_down_textual/20240826-213221_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1057MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/008_pudding_box_google_16k/008_pudding_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/008_upside_down_textual/20240826-213145_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1058MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/008_pudding_box_google_16k/008_pudding_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/008_upside_down_textual/20240826-213506_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1059MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/008_pudding_box_google_16k/008_pudding_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/008_upside_down_textual/20240826-224153_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1060MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path="data_isaaclab/assets/open6dor/ycb_16k_backup/008_pudding_box_google_16k/008_pudding_box_google_16k.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/008_upside_down_textual/20240826-224304_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1276MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9660e0c0326b4f7386014e27717231ae/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9660e0c0326b4f7386014e27717231ae_upside_down_textual/20240826-215851_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1277MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9660e0c0326b4f7386014e27717231ae/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9660e0c0326b4f7386014e27717231ae_upside_down_textual/20240826-222800_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1278MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9660e0c0326b4f7386014e27717231ae/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9660e0c0326b4f7386014e27717231ae_upside_down_textual/20240826-215949_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1279MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9660e0c0326b4f7386014e27717231ae/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9660e0c0326b4f7386014e27717231ae_upside_down_textual/20240826-224406_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1280MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9660e0c0326b4f7386014e27717231ae/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9660e0c0326b4f7386014e27717231ae_upside_down_textual/20240826-224115_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUpsideDownTextualBox1281MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="box",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/9660e0c0326b4f7386014e27717231ae/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/9660e0c0326b4f7386014e27717231ae_upside_down_textual/20240826-213244_no_interaction/trajectory-unified_wo_traj_v2.pkl"



from __future__ import annotations

from loguru import logger as log

from metasim.cfg.objects import (
    ArticulationObjMetaCfg,
    BaseObjMetaCfg,
    PrimitiveCubeMetaCfg,
    PrimitiveSphereMetaCfg,
    RigidObjMetaCfg,
)
from metasim.cfg.robots.base_robot_metacfg import BaseRobotMetaCfg
from metasim.cfg.scenario import ScenarioMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.types import EnvState, Extra, Obs, Reward, Success, TimeOut
from metasim.utils import to_snake_case

from ..base import BaseSimHandler, EnvWrapper, IdentityEnvWrapper

pass
################################################################################
## Common imports ends here, below are simulator specific imports
################################################################################

import math

import numpy as np

from isaacgym import gymapi, gymtorch, gymutil  # noqa: F401

pass
import torch


class IsaacgymHandler(BaseSimHandler):
    def __init__(self, scenario: ScenarioMetaCfg, num_envs: int = 1, headless: bool = False):
        super().__init__(scenario, num_envs, headless)
        self._task: BaseTaskMetaCfg = scenario.task
        self._robot: BaseRobotMetaCfg = scenario.robot
        self._robot_init_pose = (0, 0, 0) if not self._robot.default_position else self._robot.default_position

        self._cameras = scenario.cameras

        self._headless = headless

        self.gym = None
        self.sim = None
        self.viewer = None
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self._num_envs: int = num_envs
        self._episode_length_buf = [0 for _ in range(num_envs)]

        # asset related
        self._asset_dict_dict: dict = {}  # dict of object link index dict
        self._articulated_asset_dict_dict: dict = {}  # dict of articulated object link index dict
        self._articulated_joint_dict_dict: dict = {}  # dict of articulated object joint index dict
        self._robot_link_dict: dict = {}  # dict of robot link index dict
        self._robot_joint_dict: dict = {}  # dict of robot joint index dict
        self._joint_info: dict = {}  # dict of joint names of each env
        self._num_joints: int = 0
        self._body_info: dict = {}  # dict of body names of each env
        self._num_bodies: int = 0

        # environment related pointers
        self._envs: list = []
        self._obj_handles: list = []  # 2 dim list: list in list, each list contains object handles of each env
        self._articulated_obj_handles: list = []  # 2 dim list: list in list, each list contains articulated object handles of each env
        self._robot_handles: list = []  # 2 dim list: list of robot handles of each env

        # environment related tensor indices
        self._env_rigid_body_global_indices: list = []  # 2 dim list: list in list, each list contains global indices of each env

        self._root_states_Tensor = None  # will update after refresh
        self._dof_states_Tensor = None  # will update after refresh
        self._rigid_body_states_Tensor = None  # will update after refresh

    def launch(self) -> None:
        super().launch()
        TaskName = self._task.__class__.__name__.replace("MetaCfg", "")
        task_name = to_snake_case(TaskName)

        ## IsaacGym Initialization
        self._init_gym()
        self._make_envs()
        self._set_up_camera()
        # ==== prepare tensors =====
        # from now on, we will use the tensor API that can run on CPU or GPU
        self.gym.prepare_sim(self.sim)
        self._root_states_Tensor = gymtorch.wrap_tensor(self.gym.acquire_actor_root_state_tensor(self.sim))
        self._dof_states_Tensor = gymtorch.wrap_tensor(self.gym.acquire_dof_state_tensor(self.sim))
        self._rigid_body_states_Tensor = gymtorch.wrap_tensor(self.gym.acquire_rigid_body_state_tensor(self.sim))

    def _init_gym(self) -> None:
        physics_engine = gymapi.SIM_PHYSX
        self.gym = gymapi.acquire_gym()
        # configure sim
        sim_params = gymapi.SimParams()
        sim_params.up_axis = gymapi.UP_AXIS_Z
        sim_params.gravity = gymapi.Vec3(0.0, 0.0, -9.8)
        sim_params.dt = 1.0 / 60.0
        sim_params.substeps = 2
        sim_params.use_gpu_pipeline = True
        sim_params.physx.solver_type = 1
        sim_params.physx.num_position_iterations = 8
        sim_params.physx.num_velocity_iterations = 1
        sim_params.physx.rest_offset = 0.0
        sim_params.physx.contact_offset = 0.001
        sim_params.physx.friction_offset_threshold = 0.001
        sim_params.physx.friction_correlation_distance = 0.0005
        sim_params.physx.num_threads = 0
        sim_params.physx.use_gpu = True

        compute_device_id = 0
        graphics_device_id = 0
        self.sim = self.gym.create_sim(compute_device_id, graphics_device_id, physics_engine, sim_params)
        if self.sim is None:
            raise Exception("Failed to create sim")
        if not self._headless:
            self.viewer = self.gym.create_viewer(self.sim, gymapi.CameraProperties())
            if self.viewer is None:
                raise Exception("Failed to create viewer")

    def _set_up_camera(self) -> None:
        self._depth_tensors = []
        self._rgb_tensors = []
        self._seg_tensors = []
        self._vinv_mats = []
        self._proj_mats = []
        self._camera_handles = []
        self._env_origin = []
        for i_env in range(self.num_envs):
            self._depth_tensors.append([])
            self._rgb_tensors.append([])
            self._seg_tensors.append([])
            self._vinv_mats.append([])
            self._proj_mats.append([])
            self._env_origin.append([])

            origin = self.gym.get_env_origin(self._envs[i_env])
            self._env_origin[i_env] = [origin.x, origin.y, origin.z]
            for cam_cfg in self._cameras:
                camera_props = gymapi.CameraProperties()
                camera_props.width = cam_cfg.width
                camera_props.height = cam_cfg.height
                camera_props.horizontal_fov = cam_cfg.horizontal_fov
                camera_props.near_plane = cam_cfg.clipping_range[0]
                camera_props.far_plane = cam_cfg.clipping_range[1]
                camera_props.enable_tensors = True
                camera_handle = self.gym.create_camera_sensor(self._envs[i_env], camera_props)
                self._camera_handles.append(camera_handle)

                camera_eye = gymapi.Vec3(*cam_cfg.pos)
                camera_lookat = gymapi.Vec3(*cam_cfg.look_at)
                self.gym.set_camera_location(camera_handle, self._envs[i_env], camera_eye, camera_lookat)

                camera_tensor_depth = self.gym.get_camera_image_gpu_tensor(
                    self.sim, self._envs[i_env], camera_handle, gymapi.IMAGE_DEPTH
                )
                camera_tensor_rgb = self.gym.get_camera_image_gpu_tensor(
                    self.sim, self._envs[i_env], camera_handle, gymapi.IMAGE_COLOR
                )
                camera_tensor_rgb_seg = self.gym.get_camera_image_gpu_tensor(
                    self.sim, self._envs[i_env], camera_handle, gymapi.IMAGE_SEGMENTATION
                )
                torch_cam_depth_tensor = gymtorch.wrap_tensor(camera_tensor_depth)
                torch_cam_rgb_tensor = gymtorch.wrap_tensor(camera_tensor_rgb)
                torch_cam_rgb_seg_tensor = gymtorch.wrap_tensor(camera_tensor_rgb_seg)

                cam_vinv = torch.inverse(
                    torch.tensor(self.gym.get_camera_view_matrix(self.sim, self._envs[i_env], camera_handle))
                ).to(self.device)
                cam_proj = torch.tensor(
                    self.gym.get_camera_proj_matrix(self.sim, self._envs[i_env], camera_handle),
                    device=self.device,
                )

                self._depth_tensors[i_env].append(torch_cam_depth_tensor)
                self._rgb_tensors[i_env].append(torch_cam_rgb_tensor)
                self._seg_tensors[i_env].append(torch_cam_rgb_seg_tensor)
                self._vinv_mats[i_env].append(cam_vinv)
                self._proj_mats[i_env].append(cam_proj)

    def _load_object_asset(self, object: BaseObjMetaCfg) -> None:
        asset_root = "."
        if isinstance(object, PrimitiveCubeMetaCfg):
            asset_options = gymapi.AssetOptions()
            asset_options.armature = 0.01
            asset_options.fix_base_link = False
            asset_options.disable_gravity = False
            asset_options.flip_visual_attachments = False
            asset = self.gym.create_box(self.sim, object.size[0], object.size[1], object.size[2], asset_options)
        elif isinstance(object, ArticulationObjMetaCfg):
            asset_path = object.urdf_path
            asset_options = gymapi.AssetOptions()
            asset_options.armature = 0.01
            asset_options.fix_base_link = True
            asset_options.disable_gravity = False
            asset_options.flip_visual_attachments = False
            asset = self.gym.load_asset(self.sim, asset_root, asset_path, asset_options)
            self._articulated_asset_dict_dict[object.name] = self.gym.get_asset_rigid_body_dict(asset)
            self._articulated_joint_dict_dict[object.name] = self.gym.get_asset_dof_dict(asset)
        elif isinstance(object, RigidObjMetaCfg):
            asset_path = object.urdf_path
            asset_options = gymapi.AssetOptions()
            asset_options.armature = 0.01
            asset_options.fix_base_link = False
            asset_options.disable_gravity = False
            asset_options.flip_visual_attachments = False
            asset = self.gym.load_asset(self.sim, asset_root, asset_path, asset_options)

        asset_link_dict = self.gym.get_asset_rigid_body_dict(asset)
        self._asset_dict_dict[object.name] = asset_link_dict
        return asset

    def _load_robot_assets(self) -> None:
        asset_root = "."
        robot_asset_file = self._robot.urdf_path
        asset_options = gymapi.AssetOptions()
        asset_options.armature = 0.01
        asset_options.fix_base_link = self._robot.fix_base_link
        asset_options.disable_gravity = not self._robot.enabled_gravity
        asset_options.flip_visual_attachments = self._robot.isaacgym_flip_visual_attachments
        robot_asset = self.gym.load_asset(self.sim, asset_root, robot_asset_file, asset_options)
        # configure robot dofs
        robot_dof_props = self.gym.get_asset_dof_properties(robot_asset)
        robot_lower_limits = robot_dof_props["lower"]
        robot_upper_limits = robot_dof_props["upper"]
        robot_ranges = robot_upper_limits - robot_lower_limits
        robot_mids = 0.3 * (robot_upper_limits + robot_lower_limits)
        robot_dof_props["driveMode"][:7].fill(gymapi.DOF_MODE_POS)
        robot_dof_props["stiffness"][:7].fill(400.0)
        robot_dof_props["damping"][:7].fill(40.0)

        # grippers
        robot_dof_props["driveMode"][7:].fill(gymapi.DOF_MODE_POS)
        robot_dof_props["stiffness"][7:].fill(800.0)
        robot_dof_props["damping"][7:].fill(40.0)

        robot_num_dofs = self.gym.get_asset_dof_count(robot_asset)
        default_dof_pos = np.zeros(robot_num_dofs, dtype=np.float32)
        default_dof_pos[:7] = robot_mids[:7]
        # grippers open
        default_dof_pos[7:] = robot_upper_limits[7:]

        default_dof_state = np.zeros(robot_num_dofs, gymapi.DofState.dtype)
        default_dof_state["pos"] = default_dof_pos

        # # get link index of panda hand, which we will use as end effector
        self._robot_link_dict = self.gym.get_asset_rigid_body_dict(robot_asset)
        self._robot_joint_dict = self.gym.get_asset_dof_dict(robot_asset)

        return robot_asset, robot_dof_props, default_dof_state, default_dof_pos

    def _make_envs(
        self,
    ) -> None:
        # configure env grid
        num_per_row = int(math.sqrt(self.num_envs))
        spacing = 1.0
        env_lower = gymapi.Vec3(-spacing, -spacing, 0.0)
        env_upper = gymapi.Vec3(spacing, spacing, spacing)
        log.info("Creating %d environments" % self.num_envs)

        robot_pose = gymapi.Transform()
        robot_pose.p = gymapi.Vec3(*self._robot_init_pose)

        # add ground plane
        plane_params = gymapi.PlaneParams()
        plane_params.normal = gymapi.Vec3(0, 0, 1)
        self.gym.add_ground(self.sim, plane_params)

        # get object and robot asset
        obj_assets_list = [self._load_object_asset(obj) for obj in self._task.objects]
        # box_asset = self._load_object_asset(task.objects[0])
        robot_asset, robot_dof_props, default_dof_state, default_dof_pos = self._load_robot_assets()

        #### Joint Info ####
        for art_obj_name, art_obj_joint_dict in self._articulated_joint_dict_dict.items():
            num_joints = len(art_obj_joint_dict)
            joint_names_ = []
            for joint_i in range(num_joints):
                for joint_name, joint_idx in art_obj_joint_dict.items():
                    if joint_idx == joint_i:
                        joint_names_.append(joint_name)
            assert len(joint_names_) == num_joints
            joint_info_ = {}
            joint_info_["names"] = joint_names_
            joint_info_["local_indices"] = art_obj_joint_dict
            art_obj_joint_dict_global = {k_: v_ + self._num_joints for k_, v_ in art_obj_joint_dict.items()}
            joint_info_["global_indices"] = art_obj_joint_dict_global
            self._num_joints += num_joints
            self._joint_info[art_obj_name] = joint_info_

        # robot
        num_joints = len(self._robot_joint_dict)
        joint_names_ = []
        for joint_i in range(num_joints):
            for joint_name, joint_idx in self._robot_joint_dict.items():
                if joint_idx == joint_i:
                    joint_names_.append(joint_name)

        assert len(joint_names_) == num_joints
        joint_info_ = {}
        joint_info_["names"] = joint_names_
        joint_info_["local_indices"] = self._robot_joint_dict
        joint_info_["global_indices"] = {k_: v_ + self._num_joints for k_, v_ in self._robot_joint_dict.items()}
        self._joint_info[self._robot.name] = joint_info_
        self._num_joints += num_joints

        ###################
        #### Body Info ####
        for obj_name, asset_dict in self._asset_dict_dict.items():
            num_bodies = len(asset_dict)
            rigid_body_names = []
            for i in range(num_bodies):
                for rigid_body_name, rigid_body_idx in asset_dict.items():
                    if rigid_body_idx == i:
                        rigid_body_names.append(rigid_body_name)
            assert len(rigid_body_names) == num_bodies
            body_info_ = {}
            body_info_["name"] = rigid_body_names
            body_info_["local_indices"] = asset_dict
            body_info_["global_indices"] = {k_: v_ + self._num_bodies for k_, v_ in asset_dict.items()}
            self._body_info[obj_name] = body_info_
            self._num_bodies += num_bodies

        num_bodies = len(self._robot_link_dict)
        rigid_body_names = []
        for i in range(num_bodies):
            for rigid_body_name, rigid_body_idx in self._robot_link_dict.items():
                if rigid_body_idx == i:
                    rigid_body_names.append(rigid_body_name)

        assert len(rigid_body_names) == num_bodies
        rigid_body_info_ = {}
        rigid_body_info_["name"] = rigid_body_names
        rigid_body_info_["local_indices"] = self._robot_link_dict
        rigid_body_info_["global_indices"] = {k_: v_ + self._num_bodies for k_, v_ in self._robot_link_dict.items()}
        self._body_info[self._robot.name] = rigid_body_info_
        self._num_bodies += num_bodies

        #################

        for i in range(self.num_envs):
            # create env
            env = self.gym.create_env(self.sim, env_lower, env_upper, num_per_row)

            ##  state update  ##
            self._envs.append(env)
            self._obj_handles.append([])
            self._env_rigid_body_global_indices.append({})
            self._articulated_obj_handles.append([])
            ####################

            # carefully set each object
            for obj_i, obj_asset in enumerate(obj_assets_list):
                # add object
                obj_pose = gymapi.Transform()
                obj_pose.p.x = obj_i * 0.2  # place to any position, will update immediately at reset stage
                obj_pose.p.y = 0.0
                obj_pose.p.z = 0.0
                obj_pose.r = gymapi.Quat.from_axis_angle(gymapi.Vec3(0, 0, 1), 0)
                obj_handle = self.gym.create_actor(env, obj_asset, obj_pose, "object", i, 0)

                self.gym.set_actor_scale(env, obj_handle, self._task.objects[obj_i].scale)
                if isinstance(self._task.objects[obj_i], PrimitiveCubeMetaCfg):
                    color = gymapi.Vec3(
                        self._task.objects[obj_i].color[0],
                        self._task.objects[obj_i].color[1],
                        self._task.objects[obj_i].color[2],
                    )
                    self.gym.set_rigid_body_color(env, obj_handle, 0, gymapi.MESH_VISUAL_AND_COLLISION, color)
                elif isinstance(self._task.objects[obj_i], PrimitiveSphereMetaCfg):
                    raise NotImplementedError
                elif isinstance(self._task.objects[obj_i], RigidObjMetaCfg):
                    pass
                elif isinstance(self._task.objects[obj_i], ArticulationObjMetaCfg):
                    self._articulated_obj_handles[-1].append(obj_handle)
                else:
                    log.error("Unknown object type")
                    raise NotImplementedError
                self._obj_handles[-1].append(obj_handle)

                object_rigid_body_indices = {}
                for rigid_body_name, rigid_body_idx in self._asset_dict_dict[self._task.objects[obj_i].name].items():
                    rigid_body_idx = self.gym.find_actor_rigid_body_index(
                        env, obj_handle, rigid_body_name, gymapi.DOMAIN_SIM
                    )
                    object_rigid_body_indices[rigid_body_name] = rigid_body_idx

                self._env_rigid_body_global_indices[-1][self._task.objects[obj_i].name] = object_rigid_body_indices

            # # carefully add robot
            robot_handle = self.gym.create_actor(env, robot_asset, robot_pose, "robot", i, 2)
            self.gym.set_actor_scale(env, robot_handle, self._robot.scale)
            self._robot_handles.append(robot_handle)
            # set dof properties
            self.gym.set_actor_dof_properties(env, robot_handle, robot_dof_props)

            robot_rigid_body_indices = {}
            for rigid_body_name, rigid_body_idx in self._robot_link_dict.items():
                rigid_body_idx = self.gym.find_actor_rigid_body_index(
                    env, robot_handle, rigid_body_name, gymapi.DOMAIN_SIM
                )
                robot_rigid_body_indices[rigid_body_name] = rigid_body_idx

            self._env_rigid_body_global_indices[-1]["robot"] = robot_rigid_body_indices

        # GET initial state, copy for reset later
        self._initial_state = np.copy(self.gym.get_sim_rigid_body_states(self.sim, gymapi.STATE_ALL))

        ###### set VEWIER camera ######
        # point camera at middle env
        if not self._headless:  # TODO: update a default viewer
            cam_pos = gymapi.Vec3(1, 1, 1)
            cam_target = gymapi.Vec3(-1, -1, -0.5)
            middle_env = self._envs[self.num_envs // 2 + num_per_row // 2]
            self.gym.viewer_camera_look_at(self.viewer, middle_env, cam_pos, cam_target)
        ################################

    def get_states(self, env_ids: list[int] | None = None) -> list[EnvState]:
        if env_ids is None:
            env_ids = list(range(self.num_envs))

        states = []
        for env_id in env_ids:
            env_state = {}
            for obj_id, obj in enumerate(self._task.objects + [self._robot]):
                obj_state = {}

                ## Basic states
                obj_state["pos"] = self._root_states_Tensor[obj_id, :3].cpu()
                obj_state["rot"] = self._root_states_Tensor[obj_id, 3:7].cpu()
                obj_state["vel"] = self._root_states_Tensor[obj_id, 7:10].cpu()
                obj_state["ang_vel"] = self._root_states_Tensor[obj_id, 10:].cpu()

                ## Joint states
                if isinstance(obj, ArticulationObjMetaCfg):
                    obj_state["dof_pos"] = {
                        joint_name: self._dof_states_Tensor.reshape(
                            self.num_envs, -1, self._dof_states_Tensor.shape[1]
                        )[:, global_idx, 0]
                        for joint_name, global_idx in (self._joint_info[obj.name]["global_indices"]).items()
                    }
                    obj_state["dof_vel"] = {
                        joint_name: self._dof_states_Tensor.reshape(
                            self.num_envs, -1, self._dof_states_Tensor.shape[1]
                        )[:, global_idx, 1]
                        for joint_name, global_idx in (self._joint_info[obj.name]["global_indices"]).items()
                    }

                ## Actuator states
                ## XXX: Could non-robot objects have actuators?
                if isinstance(obj, BaseRobotMetaCfg):
                    obj_state["dof_pos_target"] = {
                        joint_name: None  # TODO
                        for i, joint_name in enumerate(self._joint_info[obj.name]["names"])
                    }
                    obj_state["dof_vel_target"] = {
                        joint_name: None  # TODO
                        for i, joint_name in enumerate(self._joint_info[obj.name]["names"])
                    }
                    obj_state["dof_torque"] = {
                        joint_name: None  # TODO
                        for i, joint_name in enumerate(self._joint_info[obj.name]["names"])
                    }
                env_state[obj.name] = obj_state

                ## Body states
                ### XXX: will have bug when there are multiple objects with same structure, e.g. dual gripper

                for i, body_name in enumerate(self._body_info[obj.name]["name"]):
                    body_state = {}
                    body_state["pos"] = self._rigid_body_states_Tensor[
                        self._body_info[obj.name]["global_indices"][body_name], :3
                    ].cpu()
                    body_state["rot"] = self._rigid_body_states_Tensor[
                        self._body_info[obj.name]["global_indices"][body_name], 3:7
                    ].cpu()
                    body_state["vel"] = self._rigid_body_states_Tensor[
                        self._body_info[obj.name]["global_indices"][body_name], 7:10
                    ].cpu()
                    body_state["ang_vel"] = self._rigid_body_states_Tensor[
                        self._body_info[obj.name]["global_indices"][body_name], 10:
                    ].cpu()
                    body_state["com"] = self._rigid_body_states_Tensor[
                        self._body_info[obj.name]["global_indices"][body_name], :3
                    ].cpu()
                    env_state[f"metasim_body_{body_name}"] = body_state

            states.append(env_state)
        return states

    def get_observation(self, action=None) -> dict:
        # images
        rgb_images = []
        depth_images = []
        seg_images = []
        for i_env in range(self.num_envs):
            for i_cam in range(len(self._cameras)):
                rgb_images.append(self._rgb_tensors[i_env][i_cam].unsqueeze(0)[..., :3])
                depth_images.append(self._depth_tensors[i_env][i_cam].unsqueeze(0))

        joint_qpos = (
            self._dof_states_Tensor.reshape(self.num_envs, -1)[:, -len(self._joint_info[self._robot.name]["names"]) :]
            .clone()
            .cpu()
        )
        if self._robot.ee_prim_path is not None:
            robot_ee_state = (
                self._rigid_body_states_Tensor.reshape(self.num_envs, -1, self._rigid_body_states_Tensor.shape[1])[
                    :, self._env_rigid_body_global_indices[0]["robot"][self._robot.ee_prim_path], :
                ]
                .clone()
                .cpu()
            )  # object first, robot last
        else:
            robot_ee_state = None

        robot_root_state = (
            self._root_states_Tensor.reshape(self.num_envs, -1, self._root_states_Tensor.shape[1])[:, -1, :]
            .clone()
            .cpu()
        )  # object first, robot last

        robot_body_state = (
            self._rigid_body_states_Tensor.reshape(self.num_envs, -1, self._rigid_body_states_Tensor.shape[1])[
                :, -len(self._env_rigid_body_global_indices[0]["robot"]) :, :
            ]
            .clone()
            .cpu()
        )
        # action
        if action is not None:
            joint_qpos_target = torch.tensor(action, dtype=torch.float32, device=self.device).reshape(self.num_envs, -1)

        else:
            joint_qpos_target = torch.zeros(self.num_envs, len(self._joint_info[self._robot.name]["names"]))

        joint_qpos_target = joint_qpos_target.reshape(self.num_envs, -1)

        if len(self._cameras) > 0:
            rgb_tensor = torch.cat(rgb_images, dim=0)
            depth_tensor = torch.cat(depth_images, dim=0)
        else:
            rgb_tensor = None
            depth_tensor = None

        data_dict = {
            "rgb": rgb_tensor,
            "depth": depth_tensor,
            ## Camera
            "cam_pos": [torch.tensor(self._cameras[i].pos) for i in range(len(self._cameras))],
            "cam_look_at": [torch.tensor(self._cameras[i].look_at) for i in range(len(self._cameras))],
            "cam_intr": [torch.zeros(3, 3) for i in range(len(self._cameras))],  # TODO: get intrinsic matrix
            "cam_extr": [torch.zeros(4, 4) for i in range(len(self._cameras))],  # TODO: get extrinsic matrix
            ## State
            "joint_qpos_target": joint_qpos_target,
            "joint_qpos": joint_qpos,  # align with old version
            "robot_ee_state": robot_ee_state,
            "robot_root_state": robot_root_state,
            "robot_body_state": robot_body_state,
        }

        for k, v in data_dict.items():
            if isinstance(v, torch.Tensor):
                data_dict[k] = v.cpu()
        torch.cuda.empty_cache()

        return data_dict

    @property
    def episode_length_buf(self) -> list[int]:
        return self._episode_length_buf

    def get_success(self) -> Success:
        return self.task.checker.check(self)

    ############################################################
    ## Gymnasium main methods
    ############################################################
    def step(self, action: list[float]) -> tuple[Obs, Reward, Success, TimeOut, Extra]:
        self._episode_length_buf = [self._episode_length_buf[i] + 1 for i in range(self.num_envs)]
        TIME_STEPS_TO_RUN = 3
        for _ in range(TIME_STEPS_TO_RUN):
            # Deploy actions

            # TODO: can only work for one env
            action_input = torch.zeros_like(self._dof_states_Tensor[:, 0].clone())

            action_array_list = []

            for action_data in action:
                flat_vals = []
                for joint_i, joint_name in enumerate(self._joint_info[self._robot.name]["names"]):
                    flat_vals.append(action_data["dof_pos_target"][joint_name])  # TODO: support other actions

                action_array = torch.tensor(flat_vals, dtype=torch.float32, device=self.device).unsqueeze(0)

                action_array_list.append(action_array)

            action_array_all = torch.cat(action_array_list, dim=0)
            robot_dim = action_array_all.shape[1]

            assert (
                action_input.shape[0] % self._num_envs == 0
            )  # WARNING: obj dim(env0), robot dim(env0), obj dim(env1), robot dim(env1) ...
            chunk_size = action_input.shape[0] // self._num_envs
            robot_dim_index = [
                i * chunk_size + offset
                for i in range(self.num_envs)
                for offset in range(chunk_size - robot_dim, chunk_size)
            ]
            action_input[robot_dim_index] = torch.tensor(
                action_array_all, dtype=torch.float32, device=self.device
            ).reshape(-1)

            if _ == 0:
                self.gym.set_dof_position_target_tensor(self.sim, gymtorch.unwrap_tensor(action_input))

            # step the physics
            self.gym.simulate(self.sim)
            self.gym.fetch_results(self.sim, True)

            # refresh tensors
            self.gym.refresh_rigid_body_state_tensor(self.sim)
            self.gym.refresh_actor_root_state_tensor(self.sim)
            self.gym.refresh_dof_state_tensor(self.sim)
            self.gym.refresh_jacobian_tensors(self.sim)
            self.gym.refresh_mass_matrix_tensors(self.sim)

            self.gym.render_all_camera_sensors(self.sim)

            # update viewer
            self.gym.step_graphics(self.sim)
            if not self._headless:
                self.gym.draw_viewer(self.viewer, self.sim, False)
            self.gym.sync_frame_time(self.sim)

            # obs
            self.gym.start_access_image_tensors(self.sim)
            obs = self.get_observation(action_array_all)
            self.gym.end_access_image_tensors(self.sim)

        # log.warning("Success, Timeout is not implemented")

        success = self.get_success()
        # reward = self.get_reward()
        reward = None

        return (
            obs,
            reward,
            success,
            torch.zeros(self.num_envs, dtype=torch.bool),
            {},
        )  # obs, reward, success, time_out, extras

    def reset(self, env_ids: list[int] | None = None) -> tuple[Obs, Extra]:
        if env_ids is None:
            env_ids = list(range(self.num_envs))

        # Reset episode length buffer
        self._episode_length_buf = [0 for _ in range(self.num_envs)]

        # Step the physics to apply the reset
        self.gym.simulate(self.sim)
        self.gym.fetch_results(self.sim, True)

        # Refresh tensors
        self.gym.refresh_rigid_body_state_tensor(self.sim)
        self.gym.refresh_actor_root_state_tensor(self.sim)
        self.gym.refresh_dof_state_tensor(self.sim)
        self.gym.refresh_jacobian_tensors(self.sim)
        self.gym.refresh_mass_matrix_tensors(self.sim)

        self.gym.render_all_camera_sensors(self.sim)

        # Update viewer
        self.gym.step_graphics(self.sim)
        if not self._headless:
            self.gym.draw_viewer(self.viewer, self.sim, False)
        self.gym.sync_frame_time(self.sim)

        # Call task checker reset
        self.task.checker.reset(self, env_ids=env_ids)

        # Get observation
        self.gym.start_access_image_tensors(self.sim)
        obs = self.get_observation()
        self.gym.end_access_image_tensors(self.sim)

        return obs, None

    def set_states(self, init_states, env_ids: list[int] | None = None):
        if env_ids is None:
            env_ids = list(range(self.num_envs))
        self.init_states_to_set = init_states

        assert len(self.init_states_to_set) == self.num_envs
        init_pos_list = []
        init_rot_list = []
        init_q_list = []
        for init_state in self.init_states_to_set:
            init_pos_list_i = []
            init_rot_list_i = []
            init_q_list_i = []
            for obj in self._task.objects:
                obj_name = obj.name
                pos = np.array(init_state[obj_name]["pos"])
                init_obj_pos = pos  # IsaacGym convention
                rot = np.array(init_state[obj_name]["rot"])
                init_obj_quat = [rot[1], rot[2], rot[3], rot[0]]  # IsaacGym convention

                init_pos_list_i.append(init_obj_pos)
                init_rot_list_i.append(init_obj_quat)
                if isinstance(obj, ArticulationObjMetaCfg):
                    init_obj_joint_q = np.zeros(len(self._articulated_joint_dict_dict[obj_name]))
                    articulated_joint_dict = self._articulated_joint_dict_dict[obj_name]
                    for joint_name, joint_idx in articulated_joint_dict.items():
                        init_obj_joint_q[joint_idx] = init_state[obj_name]["dof_pos"][joint_name]
                    init_q_list_i.append(init_obj_joint_q)

            init_pos_list_i.append(np.array(init_state[self._robot.name]["pos"]))
            rot = np.array(init_state[self._robot.name]["rot"])
            init_robot_quat = [rot[1], rot[2], rot[3], rot[0]]
            init_rot_list_i.append(init_robot_quat)

            init_robot_dof_state_i = np.zeros(len(self._robot_joint_dict))
            for joint_name, joint_idx in self._robot_joint_dict.items():
                # pdb.set_trace()
                init_robot_dof_state_i[joint_idx] = init_state[self._robot.name]["dof_pos"][joint_name]

            init_q_list_i.append(init_robot_dof_state_i)
            init_pos_list.append(init_pos_list_i)
            init_rot_list.append(init_rot_list_i)
            init_q_list.append(init_q_list_i)

        self._set_actor_root_state(init_pos_list, init_rot_list)
        self._set_actor_joint_state(init_q_list)
        # step the physics
        self.gym.simulate(self.sim)
        self.gym.fetch_results(self.sim, True)

        self.gym.refresh_actor_root_state_tensor(self.sim)
        self.gym.refresh_dof_state_tensor(self.sim)
        self.gym.refresh_rigid_body_state_tensor(self.sim)
        self.gym.refresh_jacobian_tensors(self.sim)
        self.gym.refresh_mass_matrix_tensors(self.sim)

        self.gym.render_all_camera_sensors(self.sim)

        # update viewer
        self.gym.step_graphics(self.sim)
        if not self._headless:
            self.gym.draw_viewer(self.viewer, self.sim, False)
        self.gym.sync_frame_time(self.sim)

        self.task.checker.reset(self, env_ids=env_ids)

    def _set_actor_root_state(self, position_list, rotation_list):
        new_root_states = self._root_states_Tensor.clone()
        new_root_states_pos = torch.tensor(np.array(position_list), dtype=torch.float32, device=self.device).reshape(
            -1, 3
        )
        new_root_states_rot = torch.tensor(np.array(rotation_list), dtype=torch.float32, device=self.device).reshape(
            -1, 4
        )
        new_root_states[:, :3] = new_root_states_pos
        new_root_states[:, 3:7] = new_root_states_rot
        # self.rb_states[:, self.actor_id, :7] = target_pose
        root_reset_actors_indices = torch.unique(
            torch.tensor(
                np.arange(new_root_states.shape[0]),
                dtype=torch.float32,
                device=self.device,
            )
        ).to(dtype=torch.int32)
        res = self.gym.set_actor_root_state_tensor_indexed(
            self.sim,
            gymtorch.unwrap_tensor(new_root_states),
            gymtorch.unwrap_tensor(root_reset_actors_indices),
            len(root_reset_actors_indices),
        )
        assert res

        return

    def _set_actor_joint_state(self, joint_pos_list):
        new_dof_states = self._dof_states_Tensor.clone()

        new_dof_pos = []
        # Process each element in M x N
        for env_i in range(self.num_envs):
            flat_vals = []
            for i in range(len(joint_pos_list[env_i])):
                flat_vals.extend(joint_pos_list[env_i][i])
            flat_array = np.array(flat_vals)
            new_dof_pos.append(flat_array)
        new_dof_pos = torch.tensor(np.array(new_dof_pos), dtype=torch.float32, device=self.device)

        zero_vel = torch.zeros_like(new_dof_pos)
        new_dof_states[:, 0] = new_dof_pos.reshape(-1)
        new_dof_states[:, 1] = zero_vel.reshape(-1)
        res = self.gym.set_dof_state_tensor(self.sim, gymtorch.unwrap_tensor(new_dof_states))
        assert res
        return

    def render(self) -> None:
        # update viewer
        self.gym.step_graphics(self.sim)
        if not self._headless:
            self.gym.draw_viewer(self.viewer, self.sim, False)
            self.gym.sync_frame_time(self.sim)

    def close(self) -> None:
        super().close()

    ############################################################
    ## Set object states per step
    ############################################################
    def _set_object_joint_pos(
        self,
        object: BaseObjMetaCfg,
        joint_pos: list[float],
    ) -> None:
        super()._set_object_joint_pos(object, joint_pos)
        assert len(joint_pos) == self.env.scene.articulations[object.name].num_joints

        pos = torch.tensor(joint_pos, device=self.env.device)
        vel = torch.zeros_like(pos)
        self.env.scene.articulations[object.name].write_joint_state_to_sim(
            pos, vel, env_ids=torch.tensor([0], device=self.env.device)
        )

    ############################################################
    ## Utils
    ############################################################
    def is_running(self) -> bool:
        return True
        super().is_running()

    def get_object_joint_names(self, object: BaseObjMetaCfg) -> list[str]:
        super().get_object_joint_names(object)
        return self.env.scene.articulations[object.name].joint_names

    def get_reward(self) -> Reward:
        from metasim.cfg.tasks.base_task_metacfg import BaseRLTaskMetaCfg

        if hasattr(self.task, "reward_fn"):
            return self.task.reward_fn(self.get_states())

        elif isinstance(self.task, BaseRLTaskMetaCfg):
            final_reward = torch.zeros(self.num_envs)
            for reward_func, reward_weight in zip(self.task.reward_functions, self.task.reward_weights):
                final_reward += reward_func(self.get_states()) * reward_weight
            return final_reward
        else:
            return None

    @property
    def num_envs(self) -> int:
        return self._num_envs


# TODO: try to align handler API and use GymWrapper instead
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)
IsaacgymEnv: type[EnvWrapper[IsaacgymHandler]] = IdentityEnvWrapper(IsaacgymHandler)



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class PlugChargerInPowerSupplyMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/plug_charger_in_power_supply/v2"
    objects = [
        RigidObjMetaCfg(
            name="charger",
            filepath="roboverse_data/assets/rlbench/plug_charger_in_power_supply/charger/usd/charger.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="task_wall",
            filepath="roboverse_data/assets/rlbench/plug_charger_in_power_supply/task_wall/usd/task_wall.usd",
            physics=PhysicStateType.GEOM,
        ),
        RigidObjMetaCfg(
            name="plug",
            filepath="roboverse_data/assets/rlbench/plug_charger_in_power_supply/plug/usd/plug.usd",
            physics=PhysicStateType.GEOM,
        ),
    ]
    # TODO: add checker



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class OpenJarMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/open_jar/v2"
    objects = [
        RigidObjMetaCfg(
            name="jar0",
            filepath="roboverse_data/assets/rlbench/open_jar/jar0/usd/jar0.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="jar1",
            filepath="roboverse_data/assets/rlbench/open_jar/jar1/usd/jar1.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="jar_lid0",
            filepath="roboverse_data/assets/rlbench/open_jar/jar_lid0/usd/jar_lid0.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="jar_lid1",
            filepath="roboverse_data/assets/rlbench/open_jar/jar_lid1/usd/jar_lid1.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="target",
            filepath="roboverse_data/assets/rlbench/open_jar/target/usd/target.usd",
            physics=PhysicStateType.GEOM,
        ),
    ]
    # TODO: add checker
